import java.math.BigInteger;
import java.util.Random;

/**
 * Klasse RSA.
 * 
 * @author claus albowski
 * @version okt'07
 */
public class RSA
{
    private Random random;
    int bitLength;
    private BigInteger p, q, n, phi, d, e, original, chiffrat, dechiffrat;
    private static BigInteger Big2=new BigInteger("2"); 
    

    /**
     * Konstruktor fr Objekte der Klasse RSA.
     */
    public RSA()   {
        original = new BigInteger("123456789");
        chiffrat = null;
        dechiffrat = null;
        random=new Random();
        setzeKleineWerte();
    }
    
    /**
     * Konstruktor fr Objekte der Klasse RSA
     * mit bergebener bit - Lnge. 
     */
    public RSA(int bitLaenge)   {
        setzeOriginal("123456789123456789");
        chiffrat = new BigInteger("0");
        dechiffrat = new BigInteger("0");
        random=new Random();
        bitLength=bitLaenge;
        definiereNPQPHI();
        setzeEundD();
    }
    
    /**
     * setzt das Original als BigInteger.
     */
    public void setzeOriginal(String botschaft){
        original=new BigInteger(botschaft);
    }
    
    /**
     * setzt das Chiffrat als BigInteger.
     */
    public void setzeChiffrat(String chiffratString){
        chiffrat = new BigInteger(chiffratString);
    }
    
    /**
     * zeigt das Original als String an.
     */
    public String gibOriginal(){
        return original.toString();
    }

    /**
     * zeigt das Chiffrat als String an.
     */
    public String gibChiffrat(){
        return chiffrat.toString();
    }

    /**
     * zeigt das Dechiffrat als String an.
     */
    public String gibDechiffrat(){
        return dechiffrat.toString();
    }
    
    // bersetzungsmethoden:
    
    /**
     * Potenzmodulo.
     * entspricht der Funktion modPow
     */
    public BigInteger Potenzmodulo(BigInteger basis, BigInteger hochzahl, BigInteger moduloZahl){
        if (hochzahl.equals(BigInteger.ONE)) return basis.mod(moduloZahl);
        else if (hochzahl.mod(Big2).equals(BigInteger.ZERO))
        return Potenzmodulo(basis.multiply(basis).mod(moduloZahl),
                            hochzahl.divide(Big2),
                            moduloZahl).mod(moduloZahl);
        else return (basis.multiply(Potenzmodulo(basis.multiply(basis).mod(moduloZahl),
                              hochzahl.divide(Big2),
                              moduloZahl))).mod(moduloZahl);
    }
    
    /**
     * gibt eine zufllige Pseudoprimzahl mit der bergebenen Bit-Lnge zurck.
     */
    private BigInteger gibZufallsPrimzahl(int bitLaenge){
        return new BigInteger(bitLaenge, 100, random);
    }
   
    /**
     * whlt zunchst zwei (Pseudo-)Primzahlen p und q, berechnet n und phi.
     */
    public void definiereNPQPHI(){
        p=gibZufallsPrimzahl(bitLength);
        q=gibZufallsPrimzahl(bitLength);
        n=p.multiply(q);
        phi=(p.subtract(BigInteger.ONE)).multiply(q.subtract(BigInteger.ONE));
    }

    /**
     * Bestimmung des ggT und Zerlegung des ggT in eine 
     * Linearkombination der vorgegebenen Zahlen a und b.
     * ===== Zweck bei RSA ======================================
     * Berechnet den zweiten Schlssel d zum ersten e mit Hilfe des
     * Verfahrens von Euklid-Bachet a->e , b->phi(n).
     * Verwendete Parameter :
     * erste             : jeweils die erste Zahl beim Euklid.Alg.
     * zweite         : jeweils die zweite Zahl beim Euklid.Alg.
     * xFuerErste   : Linearfaktor von a fuer die erste Zahl
     * yFuerErste   : Linearfaktor von b fuer die erste Zahl
     * xFuerZweite  : Linearfaktor von a fuer die zweite Zahl
     * yFuerZweite  : Linearfaktor von a fuer die zweite Zahl
     */
    public BigInteger euklidBachet(BigInteger a, BigInteger b){
        return
          lokal(a, b, BigInteger.ONE, BigInteger.ZERO, BigInteger.ZERO, BigInteger.ONE);
    }
    /**
     * lokal wird rekursiv verwendet.
     */
    private BigInteger lokal(BigInteger erste,
                             BigInteger zweite,
                             BigInteger xFuerErste,
                             BigInteger yFuerErste,
                             BigInteger xFuerZweite,
                             BigInteger yFuerZweite){
        if (erste.equals(BigInteger.ZERO)) return xFuerZweite;
        else if (zweite.equals(BigInteger.ZERO)) return xFuerErste;
        else return 
          lokal(erste.mod(zweite),
                zweite.mod(erste),
                xFuerErste.subtract(xFuerZweite.multiply(erste.divide(zweite))),
                yFuerErste.subtract(yFuerZweite.multiply(erste.divide(zweite))),
                xFuerZweite.subtract(xFuerErste.multiply(zweite.divide(erste))),
                yFuerZweite.subtract(yFuerErste.multiply(zweite.divide(erste))));
    }
    
    /**
     * setze e.
     */
    public void setzeEundD(){
        setzeE();
        setzeD();
    }
    
    public void setzeE(){
        // Fr e wird eine P(Pseudo-)Primzahl gewhlt, die grer als n ist.
        e=BigInteger.probablePrime(2*bitLength+1, random); // lnger als phi(n)
    }
    
    public void setzeD(){
        // Korrektur gegenueber Monikas Seifferts Text:
        // Das Ergebnis von euklidBachet kann negativ sein
        // und muss daher noch modulo phi genommen werden.
        //         phi = (p - 1)*(q - 1)
        if (e!=null)
        d = phi.add(euklidBachet(e, phi)).mod(phi);
    }
    
    /**
     * setzt kleine Werte fuer eine Beispielrechnung.
     */
    public void setzeKleineWerte(){
       bitLength=11;
       p = new BigInteger("1009"); 
       q = new BigInteger("1367");
       n = new BigInteger("1379303");
       phi = new BigInteger("1376928");
       e = new BigInteger("2560291");
       d =  new BigInteger("1209739");
    }
    
    /**
     * chiffriert mit RSA
     * und setzt das chiffrat.
     */
    public BigInteger chiffriere(){
        BigInteger ergebnis = Potenzmodulo(original,e, n);
        chiffrat = ergebnis;
        return ergebnis;
    }
    
    /**
     * dechiffriert mit RSA
     * und setzt das dechiffrat.
     */
    public BigInteger dechiffriere(){
        BigInteger ergebnis = Potenzmodulo(chiffrat, d, n);
        dechiffrat = ergebnis;
        return ergebnis;
    }
    
}